/************************************

Compile it with:

gcc dhcp_mail.c -o dhcp_mail -lpcap

Options:

-m set target mac
-i set interface to use
-n don't become an daemon process

*************************************/

#include <stdio.h>
#include <getopt.h>
#include <pcap.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <netdb.h>
#include <stdarg.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <net/ethernet.h>
#include <netinet/udp.h>
#include <netinet/ether.h>

/* Defines to easy find header of each protocol */
#define TO_IP_HEADER sizeof(struct ether_header)
#define TO_UDP_HEADER (TO_IP_HEADER + sizeof(struct ip))
#define TO_DHCP_HEADER (TO_UDP_HEADER + sizeof(struct udphdr))

/* Defines for access to smtp server */
#define SMTP_SERVER "" /* ->>> SMTP server you wish to use, can be an hostname */
#define SMTP_PORT 25  /* ->>> Default port */
#define SMTP_USERNAME "" /* ->>> Username, must be base64 encoded ;) */
#define SMTP_PASSWORD "" /* ->>> Password, same se username base64 */
#define DEST_MAIL ""     /* ->>> Destination mail */
#define SRC_MAIL ""      /* ->>> Source mail */

/* Main log file */
#define LOG_FILE "ninesix.log"

/* Globals */
FILE *loggingfs;
time_t current_time;
char *time_string;

/* My custom made dhcp header, without options field */
struct dhcphdr {
    u_int8_t opcode; /* opcode */
    u_int8_t hwtype; /* hardware type */
    u_int8_t hwaddrlen; /* hardware address lenght */
    u_int8_t hopcount; /* hop count, used only by realy agents */
    u_int32_t transid; /* transaction id */
    u_int16_t numsec; /* number of seconds */
    u_int16_t flags; /* flags */
    u_int32_t ciaddr; /* client internet address */
    u_int32_t yiaddr; /* your internet address */
    u_int32_t siaddr; /* server internet address */
    u_int32_t giaddr; /* gateway inetnet address */
    u_int8_t chwaddr [16]; /* client hardwaare address */
    u_int8_t shost [64]; /* server host name */
    u_int8_t bfile [128]; /* boot file name */
};

/* Start of option field, 99, 130, 83, 99 */
struct magic_cookie {
    u_int8_t cook[4];
};

/* Just print the usage info */
void print_usage(FILE *fs, int exit_code)
{
    fprintf(fs, "\nSixNine -h -n -i [interface] -m [mac address]\n\n");
    fprintf(fs, "-h --help              Display this short info\n");
    fprintf(fs, "-n --no-background     Prevent the program from becoming an daemon and detaching from controling tty\n");
    fprintf(fs, "-i --interface         Select wich interface to sniff on. Default is the first avaible\n");
    fprintf(fs, "-m --mac               Select wich mac to track. In short - select the target\n\n");

    exit(exit_code);
}

/* Function to make a deamon process */
void make_back_process (void)
{
    pid_t pid_fork;
    
    if((pid_fork=fork())==-1)
    {
        fprintf(stderr, ">>Error: Could not create child process: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    
    if(pid_fork!=0)
        exit(EXIT_SUCCESS);
        
    if(setsid()<0)
    {
        fprintf(stderr, ">>Error: Could not get new sid: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
}

/* Function to send msg to socket, used when sending e-mail
   No fancy stuff, no error checking (will come in time)*/
void sendSockMsg(int *sockfd, char *msg_format, ...)
{
    va_list args;
    char *msg;
    
    va_start(args, msg_format);
    vasprintf(&msg, msg_format, args);
    va_end(args);
    
    send((*sockfd), msg, strlen(msg), 0);
    
    free(msg);
}

/* Function to send an e-mail, boring. I just copied it from my prev code.
   Don't even have the will to comment it */
int send_email (char *ip_addr)
{
    int smtpSockfd, i, j, respondCode;
    struct sockaddr_in smtpServ_addr;
    struct hostent *smtpServ_host;
    char sockRecv_buff[1024];
    
    smtpServ_host=gethostbyname(SMTP_SERVER);
    if(smtpServ_host==NULL)
        return 1;

    bzero(&smtpServ_addr, sizeof(struct sockaddr_in));
    smtpServ_addr.sin_family=AF_INET;
    smtpServ_addr.sin_port=htons(SMTP_PORT);
    smtpServ_addr.sin_addr= (*((struct in_addr *)smtpServ_host->h_addr));
    
    if((smtpSockfd=socket(AF_INET, SOCK_STREAM, 0))==-1)
        return 1;
    if(connect(smtpSockfd, (struct sockaddr *)&smtpServ_addr, sizeof(struct sockaddr))==-1)
        return 1;
    
    i=0; j=0;
    while(recv(smtpSockfd, sockRecv_buff, 1023, 0))
    {
        sockRecv_buff[3]='\0';
        respondCode=atoi(sockRecv_buff);
        
        switch(respondCode)
        {
            case 220:
                sendSockMsg(&smtpSockfd, "HELO %s\n", SMTP_SERVER);
                continue;
            break;

            case 250:
                if(i==0)
                {
                    sendSockMsg(&smtpSockfd, "AUTH LOGIN\n");
                    i++;
                    continue;
                }

                if(i==1)
                {
                    sendSockMsg(&smtpSockfd, "RCPT To: %s\n", DEST_MAIL);
                    i++;
                    continue;
                }

                if(i==2)
                {
                    sendSockMsg(&smtpSockfd, "DATA\n");
                    i++;
                    continue;
                }

                if(i==3)
                {
                    sendSockMsg(&smtpSockfd, "QUIT\n");
                    continue;
                }
            break;

            case 354:
                sendSockMsg(&smtpSockfd, "Subject: Target connected to internet\n");
                sendSockMsg(&smtpSockfd, "Target connected to internet with IP: %s\n", ip_addr);
                sendSockMsg(&smtpSockfd, "\n.\n");
            break;

            case 334:
                if(j==0)
                {
                    sendSockMsg(&smtpSockfd, "%s\n", SMTP_USERNAME);
                    j++;
                    continue;
                }
                if(j==1)
                {
                    sendSockMsg(&smtpSockfd, "%s\n", SMTP_PASSWORD);
                    j++;
                    continue;
                }
            break;

            case 235:
                    sendSockMsg(&smtpSockfd, "MAIL From: %s\n", SRC_MAIL);
                    continue;
            break;
          }
    }       
    
    close(smtpSockfd);
    return 0;
}

/* Ahh here we are. Function to parse dhcp packet header and options */
void analyze_dhcp(const u_char *packet_data)
{
    struct dhcphdr *dhcp_head;
    struct magic_cookie *mgcookie;
    int is_dhcprequest=0;
    u_int8_t *option_code, *option_lenght, *dhcp_msgtype;
    u_int32_t *request_ip;
    char *ip_addr;

    dhcp_head=(struct dhcphdr *)(packet_data + TO_DHCP_HEADER); /* Go to start of dhcp header */
    fprintf(loggingfs, "\tDHCP Transaction ID 0x%x\n", ntohl(dhcp_head->transid)); /* Grab transaction id from it */
    mgcookie=(struct magic_cookie *)(packet_data + TO_DHCP_HEADER + sizeof(struct dhcphdr)); /* Go to start of option fields */
    
    /* Ahh, dance of the pointers. Beauty ;) */
    option_code=(&mgcookie->cook[3]+1*sizeof(u_int8_t)); /* Go pass the cookie, that 4 bytes, and the first option code is there */
    while(*option_code!=255)
    {
        option_lenght=(option_code+1*sizeof(u_int8_t)); /* Grab the lenght of the option */
        if(*option_code==0x32)
             request_ip=(u_int32_t *)(option_code+2*sizeof(u_int8_t)); /* Option 0x32 is requested ip address, we wan't that info */
        if(*option_code==0x35)
        {
            dhcp_msgtype=(option_code+2*sizeof(u_int8_t)); /* Option 0x35 is for DHCP REQUEST msg, comes after DHCP DISCOVER */
            if(*dhcp_msgtype==0x3)                         /*   means, host has accepted server's ip addr */
                is_dhcprequest=1;
        }
        
        option_code=(option_code+(((*option_lenght)+2)*sizeof(u_int8_t))); /* Jump to next optin, lenght of current + 2 more bytes */
    }
    if(is_dhcprequest==1)
    {
        /* We have a dhcp request msg, send e-mail to our user */
        fprintf(loggingfs, "\t\tDHCP request packet type found.\n");
        fprintf(loggingfs, "\t\tRequested IP address: %s\n", inet_ntoa(*request_ip));
        fprintf(loggingfs, "\t\tSending notification via e-mail ... ");
        asprintf(&ip_addr, "%s", inet_ntoa(*request_ip));
        send_email(ip_addr);
        free(ip_addr);
        fprintf(loggingfs, "OK\n");
    }
    else
    {
        fprintf(loggingfs, "\t\tNot an DHCP request packet\n");
    }
    
}

void analyze_packets(u_char *handler_args, const struct pcap_pkthdr *packet_header, const u_char *packet_data)
{
    /* TO KEEP IN MIND: ethernet -> IP protocol -> UDP -> DHCP */
    
    struct ether_header *eth_head;
    struct ip *ip_head;
    struct udphdr *udp_head;
    
    /* Start with ethernet header */
    eth_head=(struct ether_header *) packet_data;
    if((ntohs(eth_head->ether_type)==ETHERTYPE_IP)) /* Is IP next layer protocol? */
    {
        /* It is IP. Now let us get IP header */
        ip_head=(struct ip *)(packet_data + TO_IP_HEADER);
        if(ip_head->ip_p==IPPROTO_UDP) /* Is udp next layer (transport protocol)? */
        {
            /* It is UDP. Now let us get UDP header */
            udp_head=(struct udphdr *)(packet_data + TO_UDP_HEADER);
            if((ntohs(udp_head->source)==68) && (ntohs(udp_head->dest))) /* Check to see if DHCP is next layer protocol (using udp ports) */
            {
                current_time=time(NULL);
                asprintf(&time_string, "%s", ctime(&current_time)); /* I really hate ctime and returned string with \n on the end */
                time_string[(strlen(time_string))-1]='\0';
                fprintf(loggingfs, "\n\nFound DHCP packet [%s] ... Starting analysis\n", time_string);
                free(time_string);
                analyze_dhcp(packet_data);/* Get info from dhcp packet */
                fprintf(loggingfs, "Analysis finished\n");
                fflush(loggingfs);
            }
            
        }
    }
}

int main (int argc, char *argv[])
{
    /* Variable declaration */
    char short_options[]="hni:m:";
    struct option long_options[]={
                                  {"help",            0, NULL, 'h'},
                                  {"no-background",   0, NULL, 'n'},
                                  {"interface",       1, NULL, 'i'},
                                  {"mac",             1, NULL, 'm'},
                                  {NULL,              0, NULL,  0 }
                                 };
    int curr_opt=0, back_process, char_num;
    char *pcap_dev=NULL, pcap_errbuff[PCAP_ERRBUF_SIZE], *mac_address=NULL, *cap_filter_string;
    bpf_u_int32 dev_addr, dev_netmsk;
    struct bpf_program cap_filter;
    pcap_t *cap_fd;
    u_char *handler_args=NULL;
    /* End variable declaration */

    /* Set up default values */
    back_process=1;
    opterr=0; /* We don't want getopt_long to print err msg by itself */
    /* Done with defaults */
    
    /* Parse options */
    while(curr_opt!=-1)
    {
        curr_opt=getopt_long(argc, argv, short_options, long_options, 0);
        switch(curr_opt)
        {
            case 'h':
                print_usage(stdout, EXIT_SUCCESS);
            break;

            case 'n':
                back_process=0;
            break;

            case 'i':
                free(pcap_dev);
                char_num=strlen(optarg);
                pcap_dev=(char *)malloc(char_num+1);
                strncpy(pcap_dev, optarg, char_num);
            break;

            case 'm':
                char_num=strlen(optarg);
                mac_address=(char *)malloc(char_num+1);
                strncpy(mac_address, optarg, char_num);
            break;

            case '?':
                fprintf(stderr, ">>Error: Unknown option or missing argument\n");
                print_usage(stderr, EXIT_FAILURE);
        }
    }

    /* Daemon process ?? */
    if(back_process==1)
    {
        make_back_process();
        loggingfs=fopen(LOG_FILE, "w");
        if(loggingfs==NULL)
        {
            fprintf(stderr, ">Warning: Could not open log file, there will be no logging: %s\n", strerror(errno));
            loggingfs==stdout; /* Stdout is closed, so there will be no logging to it, it's a dummy */
        }
    }
    else
        loggingfs=stdout; /* Log to stdout if we are not deamon process */

    current_time=time(NULL);
    fprintf(loggingfs, "\n\n\nProgram started working on %s", ctime(&current_time));

    /* Set up default interface if none given */
    if(pcap_dev==NULL)
        if((pcap_dev=pcap_lookupdev(pcap_errbuff))==NULL)
        {
            fprintf(loggingfs, ">>Error: Could not find a valid device for libpcap: %s\n", pcap_errbuff);
            exit(EXIT_FAILURE);
        }
    /* No mac address? Program can't work without that */
    if(mac_address==NULL)
    {
        fprintf(loggingfs, ">>Error: No target mac address. Please select mac address\n");
        exit(EXIT_FAILURE);
    }

    /* Ok. Next cca 20 lines are just so boring initialization. Bljak. */
    if(pcap_lookupnet(pcap_dev, &dev_addr, &dev_netmsk, pcap_errbuff))
    {
        fprintf(loggingfs, ">>Error: Could not get info from interface: %s\n", pcap_errbuff);
        exit(EXIT_FAILURE);
    }

    fprintf(loggingfs, "Opening device for packet capturing ... ");
    if((cap_fd=pcap_open_live(pcap_dev, BUFSIZ, 0, -1, pcap_errbuff))==NULL)
    {
        fprintf(loggingfs, "\n>>Error: Could not open interface for packet capturing: %s\n", pcap_errbuff);
        exit(EXIT_FAILURE);
    }
    fprintf(loggingfs, "OK\n");
    
    fprintf(loggingfs, "Compiling and setting packet capturing filter ... ");
    asprintf(&cap_filter_string, "ether host %s", mac_address); /* Allocate memory to hold the filter string */
    if((pcap_compile(cap_fd, &cap_filter, cap_filter_string, 0, dev_netmsk)))
    {
        fprintf(loggingfs, "\n>>Error: Could not compile filter string\n");
        exit(EXIT_FAILURE);
    }
    free(cap_filter_string); /* Don't need that memory any more */

    if((pcap_setfilter(cap_fd, &cap_filter)))
    {
        fprintf(loggingfs, ">>Error: Could not set the filter for capturing packets: %s\n", pcap_geterr(cap_fd));
        exit(EXIT_FAILURE);
    }
    fprintf(loggingfs, "OK\n");
    /* End with main initialization */

    fprintf(loggingfs, "Starting packet capturing ...\n");
    fflush(loggingfs);
    pcap_loop(cap_fd, -1, analyze_packets, handler_args);
    fprintf(loggingfs, "Completed packet capturing!\n"); /* Ha ha :) */

    return 0;
}
